"""
 Copyright (C) 2014 Maruf Maniruzzaman
 Website: http://cosmosframework.com
 Author: Maruf Maniruzzaman
 License :: OSI Approved :: MIT License
"""

COSMOS_GLOBAL_SETTINGS_OBJECT_NAME = "cosmos.globalsettings"
COSMOS_APPLICATION_OBJECT_NAME = "cosmos.applications"
COSMOS_APPLICATION_ENDPOINT_LIST_OBJECT_NAME = "cosmos.appendpoints"
COSMOS_INTERCEPTOR_OBJECT_NAME = "cosmos.interceptors"
COSMOS_WIDGETS_OBJECT_NAME = "cosmos.widgets"
COSMOS_SOURCE_MODULES_OBJECT_NAME = "cosmos.sourcemodules"
COSMOS_SOURCE_MODULES_TYPE_EMBEDDED = "embedded"
COSMOS_SOURCE_MODULES_TYPE_GRIDFILE = "gridfile"
COSMOS_WORKFLOWS_OBJECT_NAME = "cosmos.workflows"


#Application export / import constants
COSMOS_OBJECT_DATA_FILE_NAME = "object_data.json"
COSMOS_APPLICATION_FILE_NAME = 'application.json'
COSMOS_CURRENT_ARCHIVE_VERSION = "0.01.000.00"

COSMOS_SOURCE_FILE_ROOT_NAME = "sources"